@@ -1,3 +1,4 @@
+import os
 import sys
 
 # make sure gevent-socketio is not installed, as it conflicts with
@@ -15,13 +16,18 @@ if gevent_socketio_found:
 
 import socketio
 import flask
+from flask import _request_ctx_stack, json as flask_json
+from flask.sessions import SessionMixin
 from werkzeug.debug import DebuggedApplication
 from werkzeug.serving import run_with_reloader
 
+from .namespace import Namespace
 from .test_client import SocketIOTestClient
 
+__version__ = '3.3.1'
 
-class _SocketIOMiddleware(socketio.Middleware):
+
+class _SocketIOMiddleware(socketio.WSGIApp):
     """This WSGI middleware simply exposes the Flask application in the WSGI
     environment before executing the request.
     """
@@ -29,14 +35,22 @@ class _SocketIOMiddleware(socketio.Middleware):
         self.flask_app = flask_app
         super(_SocketIOMiddleware, self).__init__(socketio_app,
                                                   flask_app.wsgi_app,
-                                                  socketio_path)
+                                                  socketio_path=socketio_path)
 
     def __call__(self, environ, start_response):
+        environ = environ.copy()
         environ['flask.app'] = self.flask_app
         return super(_SocketIOMiddleware, self).__call__(environ,
                                                          start_response)
 
 
+class _ManagedSession(dict, SessionMixin):
+    """This class is used for user sessions that are managed by
+    Flask-SocketIO. It is simple dict, expanded with the Flask session
+    attributes."""
+    pass
+
+
 class SocketIO(object):
     """Create a Flask-SocketIO server.
 
@@ -44,6 +58,14 @@ class SocketIO(object):
                 isn't known at the time this class is instantiated, then call
                 ``socketio.init_app(app)`` once the application instance is
                 available.
+    :param manage_session: If set to ``True``, this extension manages the user
+                           session for Socket.IO events. If set to ``False``,
+                           Flask's own session management is used. When using
+                           Flask's cookie based sessions it is recommended that
+                           you leave this set to the default of ``True``. When
+                           using server-side sessions, a ``False`` setting
+                           enables sharing the user session between HTTP routes
+                           and Socket.IO events.
     :param message_queue: A connection URL for a message queue service the
                           server can use for multi-process communication. A
                           message queue is not required when using a single
@@ -53,8 +75,10 @@ class SocketIO(object):
                     multiple clusters of SocketIO processes need to use the
                     same message queue without interfering with each other, then
                     each cluster should use a different channel.
-    :param resource: The SocketIO resource name. Defaults to ``'socket.io'``.
-                     Leave this as is unless you know what you are doing.
+    :param path: The path where the Socket.IO server is exposed. Defaults to
+                 ``'socket.io'``. Leave this as is unless you know what you are
+                 doing.
+    :param resource: Alias to ``path``.
     :param kwargs: Socket.IO and Engine.IO server options.
 
     The Socket.IO server options are detailed below:
@@ -66,7 +90,8 @@ class SocketIO(object):
                            cases, this argument does not need to be set
                            explicitly.
     :param logger: To enable logging set to ``True`` or pass a logger object to
-                   use. To disable logging set to ``False``.
+                   use. To disable logging set to ``False``. The default is
+                   ``False``.
     :param binary: ``True`` to support binary payloads, ``False`` to treat all
                    payloads as text. On Python 2, if this is set to ``True``,
                    ``unicode`` values are treated as text, and ``str`` and
@@ -76,74 +101,130 @@ class SocketIO(object):
     :param json: An alternative json module to use for encoding and decoding
                  packets. Custom json modules must have ``dumps`` and ``loads``
                  functions that are compatible with the standard library
-                 versions.
+                 versions. To use the same json encoder and decoder as a Flask
+                 application, use ``flask.json``.
 
     The Engine.IO server configuration supports the following settings:
 
-    :param async_mode: The library used for asynchronous operations. Valid
-                       options are "threading", "eventlet" and "gevent". If
-                       this argument is not given, "eventlet" is tried first,
-                       then "gevent", and finally "threading". The websocket
-                       transport is not supported in "ithreading" mode.
+    :param async_mode: The asynchronous model to use. See the Deployment
+                       section in the documentation for a description of the
+                       available options. Valid async modes are
+                       ``threading``, ``eventlet``, ``gevent`` and
+                       ``gevent_uwsgi``. If this argument is not given,
+                       ``eventlet`` is tried first, then ``gevent_uwsgi``,
+                       then ``gevent``, and finally ``threading``. The
+                       first async mode that has all its dependencies installed
+                       is then one that is chosen.
     :param ping_timeout: The time in seconds that the client waits for the
-                         server to respond before disconnecting.
+                         server to respond before disconnecting. The default is
+                         60 seconds.
     :param ping_interval: The interval in seconds at which the client pings
-                          the server.
+                          the server. The default is 25 seconds.
     :param max_http_buffer_size: The maximum size of a message when using the
-                                 polling transport.
-    :param allow_upgrades: Whether to allow transport upgrades or not.
+                                 polling transport. The default is 100,000,000
+                                 bytes.
+    :param allow_upgrades: Whether to allow transport upgrades or not. The
+                           default is ``True``.
     :param http_compression: Whether to compress packages when using the
-                             polling transport.
+                             polling transport. The default is ``True``.
     :param compression_threshold: Only compress messages when their byte size
-                                  is greater than this value.
+                                  is greater than this value. The default is
+                                  1024 bytes.
     :param cookie: Name of the HTTP cookie that contains the client session
                    id. If set to ``None``, a cookie is not sent to the client.
+                   The default is ``'io'``.
     :param cors_allowed_origins: List of origins that are allowed to connect
                                  to this server. All origins are allowed by
                                  default.
     :param cors_credentials: Whether credentials (cookies, authentication) are
-                             allowed in requests to this server.
+                             allowed in requests to this server. The default is
+                             ``True``.
     :param engineio_logger: To enable Engine.IO logging set to ``True`` or pass
                             a logger object to use. To disable logging set to
-                            ``False``.
+                            ``False``. The default is ``False``.
     """
 
     def __init__(self, app=None, **kwargs):
         self.server = None
-        self.server_options = None
+        self.server_options = {}
         self.wsgi_server = None
         self.handlers = []
+        self.namespace_handlers = []
         self.exception_handlers = {}
         self.default_exception_handler = None
-        if app is not None or len(kwargs) > 0:
+        self.manage_session = True
+        # We can call init_app when:
+        # - we were given the Flask app instance (standard initialization)
+        # - we were not given the app, but we were given a message_queue
+        #   (standard initialization for auxiliary process)
+        # In all other cases we collect the arguments and assume the client
+        # will call init_app from an app factory function.
+        if app is not None or 'message_queue' in kwargs:
             self.init_app(app, **kwargs)
+        else:
+            self.server_options.update(kwargs)
 
     def init_app(self, app, **kwargs):
         if app is not None:
             if not hasattr(app, 'extensions'):
                 app.extensions = {}  # pragma: no cover
             app.extensions['socketio'] = self
-        self.server_options = kwargs
+        self.server_options.update(kwargs)
+        self.manage_session = self.server_options.pop('manage_session',
+                                                      self.manage_session)
 
         if 'client_manager' not in self.server_options:
-            url = kwargs.pop('message_queue', None)
-            channel = kwargs.pop('channel', 'flask-socketio')
+            url = self.server_options.pop('message_queue', None)
+            channel = self.server_options.pop('channel', 'flask-socketio')
             write_only = app is None
             if url:
-                if url.startswith('redis://'):
+                if url.startswith(('redis://', "rediss://")):
                     queue_class = socketio.RedisManager
+                elif url.startswith('zmq'):
+                    queue_class = socketio.ZmqManager
                 else:
                     queue_class = socketio.KombuManager
                 queue = queue_class(url, channel=channel,
                                     write_only=write_only)
                 self.server_options['client_manager'] = queue
 
-        resource = kwargs.pop('resource', 'socket.io')
+        if 'json' in self.server_options and \
+                self.server_options['json'] == flask_json:
+            # flask's json module is tricky to use because its output
+            # changes when it is invoked inside or outside the app context
+            # so here to prevent any ambiguities we replace it with wrappers
+            # that ensure that the app context is always present
+            class FlaskSafeJSON(object):
+                @staticmethod
+                def dumps(*args, **kwargs):
+                    with app.app_context():
+                        return flask_json.dumps(*args, **kwargs)
+
+                @staticmethod
+                def loads(*args, **kwargs):
+                    with app.app_context():
+                        return flask_json.loads(*args, **kwargs)
+
+            self.server_options['json'] = FlaskSafeJSON
+
+        resource = self.server_options.pop('path', None) or \
+            self.server_options.pop('resource', None) or 'socket.io'
         if resource.startswith('/'):
             resource = resource[1:]
+        if os.environ.get('FLASK_RUN_FROM_CLI'):
+            if self.server_options.get('async_mode') is None:
+                if app is not None:
+                    app.logger.warning(
+                        'Flask-SocketIO is Running under Werkzeug, WebSocket '
+                        'is not available.')
+                self.server_options['async_mode'] = 'threading'
         self.server = socketio.Server(**self.server_options)
+        self.async_mode = self.server.async_mode
         for handler in self.handlers:
             self.server.on(handler[0], handler[1], namespace=handler[2])
+        for namespace_handler in self.namespace_handlers:
+            self.server.register_namespace(namespace_handler)
+
         if app is not None:
             # here we attach the SocketIO middlware to the SocketIO object so it
             # can be referenced later if debug middleware needs to be inserted
@@ -174,38 +255,14 @@ class SocketIO(object):
 
         def decorator(handler):
             def _handler(sid, *args):
-                app = self.server.environ[sid]['flask.app']
-                with app.request_context(self.server.environ[sid]):
-                    if 'saved_session' in self.server.environ[sid]:
-                        self._copy_session(
-                            self.server.environ[sid]['saved_session'],
-                            flask.session)
-                    flask.request.sid = sid
-                    flask.request.namespace = namespace
-                    flask.request.event = {'message': message, 'args': args}
-                    try:
-                        if message == 'connect':
-                            ret = handler()
-                        else:
-                            ret = handler(*args)
-                    except:
-                        err_handler = self.exception_handlers.get(
-                            namespace, self.default_exception_handler)
-                        if err_handler is None:
-                            raise
-                        type, value, traceback = sys.exc_info()
-                        return err_handler(value)
-                    if flask.session.modified:
-                        self.server.environ[sid]['saved_session'] = {}
-                        self._copy_session(
-                            flask.session,
-                            self.server.environ[sid]['saved_session'])
-                    return ret
+                return self._handle_event(handler, message, namespace, sid,
+                                          *args)
+
             if self.server:
                 self.server.on(message, _handler, namespace=namespace)
             else:
                 self.handlers.append((message, _handler, namespace))
-            return _handler
+            return handler
         return decorator
 
     def on_error(self, namespace=None):
@@ -248,6 +305,40 @@ class SocketIO(object):
         self.default_exception_handler = exception_handler
         return exception_handler
 
+    def on_event(self, message, handler, namespace=None):
+        """Register a SocketIO event handler.
+
+        ``on_event`` is the non-decorator version of ``'on'``.
+
+        Example::
+
+            def on_foo_event(json):
+                print('received json: ' + str(json))
+
+            socketio.on_event('my event', on_foo_event, namespace='/chat')
+
+        :param message: The name of the event. This is normally a user defined
+                        string, but a few event names are already defined. Use
+                        ``'message'`` to define a handler that takes a string
+                        payload, ``'json'`` to define a handler that takes a
+                        JSON blob payload, ``'connect'`` or ``'disconnect'``
+                        to create handlers for connection and disconnection
+                        events.
+        :param handler: The function that handles the event.
+        :param namespace: The namespace on which the handler is to be
+                          registered. Defaults to the global namespace.
+        """
+        self.on(message, namespace=namespace)(handler)
+
+    def on_namespace(self, namespace_handler):
+        if not isinstance(namespace_handler, Namespace):
+            raise ValueError('Not a namespace instance.')
+        namespace_handler._set_socketio(self)
+        if self.server:
+            self.server.register_namespace(namespace_handler)
+        else:
+            self.namespace_handlers.append(namespace_handler)
+
     def emit(self, event, *args, **kwargs):
         """Emit a server generated SocketIO event.
 
@@ -269,23 +360,28 @@ class SocketIO(object):
         :param room: Send the message to all the users in the given room. If
                      this parameter is not included, the event is sent to
                      all connected users.
-        :param include_self: ``True`` to include the sender when broadcasting
-                             or addressing a room, or ``False`` to send to
-                             everyone but the sender.
+        :param skip_sid: The session id of a client to ignore when broadcasting
+                         or addressing a room. This is typically set to the
+                         originator of the message, so that everyone except
+                         that client receive the message.
         :param callback: If given, this function will be called to acknowledge
                          that the client has received the message. The
                          arguments that will be passed to the function are
                          those provided by the client. Callback functions can
                          only be used when addressing an individual client.
         """
-        skip_sid = flask.request.sid \
-            if not kwargs.get('include_self', True) else None
-        self.server.emit(event, *args, namespace=kwargs.get('namespace', '/'),
-                         room=kwargs.get('room'), skip_sid=skip_sid,
-                         callback=kwargs.get('callback'))
+        namespace = kwargs.pop('namespace', '/')
+        room = kwargs.pop('room', None)
+        include_self = kwargs.pop('include_self', True)
+        skip_sid = kwargs.pop('skip_sid', None)
+        if not include_self and not skip_sid:
+            skip_sid = flask.request.sid
+        callback = kwargs.pop('callback', None)
+        self.server.emit(event, *args, namespace=namespace, room=room,
+                         skip_sid=skip_sid, callback=callback, **kwargs)
 
     def send(self, data, json=False, namespace=None, room=None,
-             callback=None, include_self=True):
+             callback=None, include_self=True, skip_sid=None, **kwargs):
         """Send a server-generated SocketIO message.
 
         This function sends a simple SocketIO message to one or more connected
@@ -294,7 +390,7 @@ class SocketIO(object):
         used outside of a SocketIO event context, so it is appropriate to use
         when the server is the originator of an event.
 
-        :param message: The message to send, either a string or a JSON blob.
+        :param data: The message to send, either a string or a JSON blob.
         :param json: ``True`` if ``message`` is a JSON blob, ``False``
                      otherwise.
         :param namespace: The namespace under which the message is to be sent.
@@ -302,22 +398,23 @@ class SocketIO(object):
         :param room: Send the message only to the users in the given room. If
                      this parameter is not included, the message is sent to
                      all connected users.
-        :param include_self: ``True`` to include the sender when broadcasting
-                             or addressing a room, or ``False`` to send to
-                             everyone but the sender.
+        :param skip_sid: The session id of a client to ignore when broadcasting
+                         or addressing a room. This is typically set to the
+                         originator of the message, so that everyone except
+                         that client receive the message.
         :param callback: If given, this function will be called to acknowledge
                          that the client has received the message. The
                          arguments that will be passed to the function are
                          those provided by the client. Callback functions can
                          only be used when addressing an individual client.
         """
-        skip_sid = flask.request.sid if not include_self else None
+        skip_sid = flask.request.sid if not include_self else skip_sid
         if json:
             self.emit('json', data, namespace=namespace, room=room,
-                      skip_sid=skip_sid, callback=callback)
+                      skip_sid=skip_sid, callback=callback, **kwargs)
         else:
             self.emit('message', data, namespace=namespace, room=room,
-                      skip_sid=skip_sid, callback=callback)
+                      skip_sid=skip_sid, callback=callback, **kwargs)
 
     def close_room(self, room, namespace=None):
         """Close a room.
@@ -408,7 +505,12 @@ class SocketIO(object):
         elif self.server.eio.async_mode == 'eventlet':
             def run_server():
                 import eventlet
-                eventlet_socket = eventlet.listen((host, port))
+                import eventlet.wsgi
+                import eventlet.green
+                addresses = eventlet.green.socket.getaddrinfo(host, port)
+                if not addresses:
+                    raise RuntimeError('Could not resolve host to a valid address')
+                eventlet_socket = eventlet.listen(addresses[0][4], addresses[0][0])
 
                 # If provided an SSL argument, use an SSL socket
                 ssl_args = ['keyfile', 'certfile', 'server_side', 'cert_reqs',
@@ -447,7 +549,7 @@ class SocketIO(object):
                     log=log, **kwargs)
             else:
                 self.wsgi_server = pywsgi.WSGIServer((host, port), app,
-                                                     log=log)
+                                                     log=log, **kwargs)
 
             if use_reloader:
                 # monkey patching is required by the reloader
@@ -477,13 +579,98 @@ class SocketIO(object):
         elif self.server.eio.async_mode == 'gevent':
             self.wsgi_server.stop()
 
-    def test_client(self, app, namespace=None):
-        """Return a simple SocketIO client that can be used for unit tests."""
-        return SocketIOTestClient(app, self, namespace)
+    def start_background_task(self, target, *args, **kwargs):
+        """Start a background task using the appropriate async model.
 
-    def _copy_session(self, src, dest):
-        for k in src:
-            dest[k] = src[k]
+        This is a utility function that applications can use to start a
+        background task using the method that is compatible with the
+        selected async mode.
+
+        :param target: the target function to execute.
+        :param args: arguments to pass to the function.
+        :param kwargs: keyword arguments to pass to the function.
+
+        This function returns an object compatible with the `Thread` class in
+        the Python standard library. The `start()` method on this object is
+        already called by this function.
+        """
+        return self.server.start_background_task(target, *args, **kwargs)
+
+    def sleep(self, seconds=0):
+        """Sleep for the requested amount of time using the appropriate async
+        model.
+
+        This is a utility function that applications can use to put a task to
+        sleep without having to worry about using the correct call for the
+        selected async mode.
+        """
+        return self.server.sleep(seconds)
+
+    def test_client(self, app, namespace=None, query_string=None,
+                    headers=None, flask_test_client=None):
+        """The Socket.IO test client is useful for testing a Flask-SocketIO
+        server. It works in a similar way to the Flask Test Client, but
+        adapted to the Socket.IO server.
+
+        :param app: The Flask application instance.
+        :param namespace: The namespace for the client. If not provided, the
+                          client connects to the server on the global
+                          namespace.
+        :param query_string: A string with custom query string arguments.
+        :param headers: A dictionary with custom HTTP headers.
+        :param flask_test_client: The instance of the Flask test client
+                                  currently in use. Passing the Flask test
+                                  client is optional, but is necessary if you
+                                  want the Flask user session and any other
+                                  cookies set in HTTP routes accessible from
+                                  Socket.IO events.
+        """
+        return SocketIOTestClient(app, self, namespace=namespace,
+                                  query_string=query_string, headers=headers,
+                                  flask_test_client=flask_test_client)
+
+    def _handle_event(self, handler, message, namespace, sid, *args):
+        if sid not in self.server.environ:
+            # we don't have record of this client, ignore this event
+            return '', 400
+        app = self.server.environ[sid]['flask.app']
+        with app.request_context(self.server.environ[sid]):
+            if self.manage_session:
+                # manage a separate session for this client's Socket.IO events
+                # created as a copy of the regular user session
+                if 'saved_session' not in self.server.environ[sid]:
+                    self.server.environ[sid]['saved_session'] = \
+                        _ManagedSession(flask.session)
+                session_obj = self.server.environ[sid]['saved_session']
+            else:
+                # let Flask handle the user session
+                # for cookie based sessions, this effectively freezes the
+                # session to its state at connection time
+                # for server-side sessions, this allows HTTP and Socket.IO to
+                # share the session, with both having read/write access to it
+                session_obj = flask.session._get_current_object()
+            _request_ctx_stack.top.session = session_obj
+            flask.request.sid = sid
+            flask.request.namespace = namespace
+            flask.request.event = {'message': message, 'args': args}
+            try:
+                if message == 'connect':
+                    ret = handler()
+                else:
+                    ret = handler(*args)
+            except:
+                err_handler = self.exception_handlers.get(
+                    namespace, self.default_exception_handler)
+                if err_handler is None:
+                    raise
+                type, value, traceback = sys.exc_info()
+                return err_handler(value)
+            if not self.manage_session:
+                # when Flask is managing the user session, it needs to save it
+                if not hasattr(session_obj, 'modified') or session_obj.modified:
+                    resp = app.response_class()
+                    app.session_interface.save_session(app, session_obj, resp)
+            return ret
 
 
 def emit(event, *args, **kwargs):
@@ -513,18 +700,30 @@ def emit(event, *args, **kwargs):
     :param include_self: ``True`` to include the sender when broadcasting or
                          addressing a room, or ``False`` to send to everyone
                          but the sender.
+    :param ignore_queue: Only used when a message queue is configured. If
+                         set to ``True``, the event is emitted to the
+                         clients directly, without going through the queue.
+                         This is more efficient, but only works when a
+                         single server process is used, or when there is a
+                         single addresee. It is recommended to always leave
+                         this parameter with its default value of ``False``.
     """
-    namespace = kwargs.get('namespace', flask.request.namespace)
+    if 'namespace' in kwargs:
+        namespace = kwargs['namespace']
+    else:
+        namespace = flask.request.namespace
     callback = kwargs.get('callback')
     broadcast = kwargs.get('broadcast')
     room = kwargs.get('room')
     if room is None and not broadcast:
         room = flask.request.sid
     include_self = kwargs.get('include_self', True)
+    ignore_queue = kwargs.get('ignore_queue', False)
 
     socketio = flask.current_app.extensions['socketio']
     return socketio.emit(event, *args, namespace=namespace, room=room,
-                         include_self=include_self, callback=callback)
+                         include_self=include_self, callback=callback,
+                         ignore_queue=ignore_queue)
 
 
 def send(message, **kwargs):
@@ -536,6 +735,8 @@ def send(message, **kwargs):
     can only be called from a SocketIO event handler.
 
     :param message: The message to send, either a string or a JSON blob.
+    :param json: ``True`` if ``message`` is a JSON blob, ``False``
+                     otherwise.
     :param namespace: The namespace under which the message is to be sent.
                       Defaults to the namespace used by the originating event.
                       An empty string can be used to use the global namespace.
@@ -548,21 +749,34 @@ def send(message, **kwargs):
     :param include_self: ``True`` to include the sender when broadcasting or
                          addressing a room, or ``False`` to send to everyone
                          but the sender.
+    :param ignore_queue: Only used when a message queue is configured. If
+                         set to ``True``, the event is emitted to the
+                         clients directly, without going through the queue.
+                         This is more efficient, but only works when a
+                         single server process is used, or when there is a
+                         single addresee. It is recommended to always leave
+                         this parameter with its default value of ``False``.
     """
-    namespace = kwargs.get('namespace', flask.request.namespace)
+    json = kwargs.get('json', False)
+    if 'namespace' in kwargs:
+        namespace = kwargs['namespace']
+    else:
+        namespace = flask.request.namespace
     callback = kwargs.get('callback')
     broadcast = kwargs.get('broadcast')
     room = kwargs.get('room')
     if room is None and not broadcast:
         room = flask.request.sid
     include_self = kwargs.get('include_self', True)
+    ignore_queue = kwargs.get('ignore_queue', False)
 
     socketio = flask.current_app.extensions['socketio']
-    return socketio.send(message, namespace=namespace, room=room,
-                         include_self=include_self, callback=callback)
+    return socketio.send(message, json=json, namespace=namespace, room=room,
+                         include_self=include_self, callback=callback,
+                         ignore_queue=ignore_queue)
 
 
-def join_room(room):
+def join_room(room, sid=None, namespace=None):
     """Join a room.
 
     This function puts the user in a room, under the current namespace. The
@@ -577,18 +791,22 @@ def join_room(room):
             send(username + ' has entered the room.', room=room)
 
     :param room: The name of the room to join.
+    :param sid: The session id of the client. If not provided, the client is
+                obtained from the request context.
+    :param namespace: The namespace for the room. If not provided, the
+                      namespace is obtained from the request context.
     """
     socketio = flask.current_app.extensions['socketio']
-    socketio.server.enter_room(flask.request.sid, room,
-                               namespace=flask.request.namespace)
+    sid = sid or flask.request.sid
+    namespace = namespace or flask.request.namespace
+    socketio.server.enter_room(sid, room, namespace=namespace)
 
 
-def leave_room(room):
+def leave_room(room, sid=None, namespace=None):
     """Leave a room.
 
     This function removes the user from a room, under the current namespace.
-    The user and the namespace are obtained from the event context. This is
-    a function that can only be called from a SocketIO event handler. Example::
+    The user and the namespace are obtained from the event context. Example::
 
         @socketio.on('leave')
         def on_leave(data):
@@ -598,38 +816,50 @@ def leave_room(room):
             send(username + ' has left the room.', room=room)
 
     :param room: The name of the room to leave.
+    :param sid: The session id of the client. If not provided, the client is
+                obtained from the request context.
+    :param namespace: The namespace for the room. If not provided, the
+                      namespace is obtained from the request context.
     """
     socketio = flask.current_app.extensions['socketio']
-    socketio.server.leave_room(flask.request.sid, room,
-                               namespace=flask.request.namespace)
+    sid = sid or flask.request.sid
+    namespace = namespace or flask.request.namespace
+    socketio.server.leave_room(sid, room, namespace=namespace)
 
 
-def close_room(room):
+def close_room(room, namespace=None):
     """Close a room.
 
     This function removes any users that are in the given room and then deletes
-    the room from the server. This is a function that can only be called from
-    a SocketIO event handler.
+    the room from the server.
 
     :param room: The name of the room to close.
+    :param namespace: The namespace for the room. If not provided, the
+                      namespace is obtained from the request context.
     """
     socketio = flask.current_app.extensions['socketio']
-    socketio.server.close_room(room, namespace=flask.request.namespace)
+    namespace = namespace or flask.request.namespace
+    socketio.server.close_room(room, namespace=namespace)
 
 
-def rooms():
+def rooms(sid=None, namespace=None):
     """Return a list of the rooms the client is in.
 
     This function returns all the rooms the client has entered, including its
-    own room, assigned by the Socket.IO server. This is a function that can
-    only be called from a SocketIO event handler.
+    own room, assigned by the Socket.IO server.
+
+    :param sid: The session id of the client. If not provided, the client is
+                obtained from the request context.
+    :param namespace: The namespace for the room. If not provided, the
+                      namespace is obtained from the request context.
     """
     socketio = flask.current_app.extensions['socketio']
-    return socketio.server.rooms(flask.request.sid,
-                                 namespace=flask.request.namespace)
+    sid = sid or flask.request.sid
+    namespace = namespace or flask.request.namespace
+    return socketio.server.rooms(sid, namespace=namespace)
 
 
-def disconnect(silent=False):
+def disconnect(sid=None, namespace=None, silent=False):
     """Disconnect the client.
 
     This function terminates the connection with the client. As a result of
@@ -639,10 +869,16 @@ def disconnect(silent=False):
         def receive_message(msg):
             if is_banned(session['username']):
                 disconnect()
-            # ...
+            else:
+                # ...
 
+    :param sid: The session id of the client. If not provided, the client is
+                obtained from the request context.
+    :param namespace: The namespace for the room. If not provided, the
+                      namespace is obtained from the request context.
     :param silent: this option is deprecated.
     """
     socketio = flask.current_app.extensions['socketio']
-    return socketio.server.disconnect(flask.request.sid,
-                                      namespace=flask.request.namespace)
+    sid = sid or flask.request.sid
+    namespace = namespace or flask.request.namespace
+    return socketio.server.disconnect(sid, namespace=namespace)
